/* sw3detect.c
 *
 * v1.00 (c) 27/04/2000 codex@bogus.net
 *
 * ++ big greetings ++   thanks to maelstrom and fake   ++ big greetingss ++
 *
 */

#define BUFSIZE 128
#define SNAPLEN 74                /* max bytes to capture (14+20+64) */ 

#ifndef DEBUG
#define DEBUG 1
#endif

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pcap/pcap.h>
#include <netinet/in.h>
#include <sys/types.h>

#ifdef LINUX
#include <linux/socket.h>
#include <linux/inet.h>
#else
#include <sys/socket.h>
#include <arpa/inet.h>
#endif

/* thanks to one pass compilations, we need these -------------------------- */

int rversion13(char*, char*);
int rversion14(char*, char*);
int dversion13(char*, char*);
int dversion14(char*, char*);
void decode_packet(char *);

/* usage() ----------------------------------------------------------------- */

void usage(void) {
      fprintf(stdout,"sw3detect v1.00 17/05/2000 (c) codex@bogus.net\n");
      fprintf(stdout,"usage: sw3detect [-h] [-i <device>]\n");
      fprintf(stdout,"\t -h : this\n");
      fprintf(stdout,"\t -i : network device (default eth0)\n");
}

/* main() -- the big kahuna ------------------------------------------------ */

int main(int argc, char *argv[]) {

  pcap_t *pd;                      /* capture device descriptor */

  char errbuf[PCAP_ERRBUF_SIZE];   /* pcap's error buffer */
  char buffer[BUFSIZE];            /* buffer to store data in */ 

  char device[256];                /* capture device */
  char *filter_spec = "icmp[0] = 0 or icmp[0] = 8";  /* echo rqst/rply */

  /* char *filter_spec = NULL;  */

  int timeout = 180;               /* read timeout in ms */
  int promisc = 1;                 /* put the i/f into promisc mode */
  int optimize = 0;                /* no optimising */

  int int_foo,c; /* dummy ints */

  unsigned char *p_ptr;            /* pointer to captured packet */

  struct pcap_pkthdr pkt_header;
  struct bpf_program filter_code;

  bpf_u_int32 network,netmask;

  bzero(device,sizeof(device));

  while((c = getopt(argc,argv,"i:h")) != EOF) {
    switch(c) {
    case 'i':
      if(sizeof(optarg)<sizeof(device)) memcpy(device,optarg,sizeof(optarg));
      break;
    case 'h':
      usage();
      exit(0);
    default:
      fprintf(stdout,"listening on eth0...\n");
    }
  }


  if(!device) memcpy(device,"eth0",4);

  /* get packet capture descriptor, exit & fail if not */

  bzero(buffer,sizeof(buffer));

  if(!(pd = pcap_open_live(device,SNAPLEN,promisc,timeout,buffer))) {
    fprintf(stderr,"Couldn't open device %s: %s (root privs?)\n",
	    device,strerror(errno));
    exit(1);
  }

  /* lookup & set network details */

  if(pcap_lookupnet(device,&network,&netmask,errbuf)<0) {
    fprintf(stderr,"Couldn't get network details: %s\n",errbuf);
    exit(1);
  };

  /* compile filter spec */

  if(pcap_compile(pd,&filter_code,filter_spec,optimize,netmask)<0) {
    fprintf(stderr,"Couldn't compile filter: %s\n",pcap_geterr(pd));
    exit(1);
  }

  /* use filter */

  if(pcap_setfilter(pd,&filter_code)<0) {
    fprintf(stderr,"Failed to set filter: %s\n",pcap_geterr(pd));
    exit(1);
  }

  /* decode detected packet buffer, do so forever */

  while(((char *)p_ptr=pcap_next(pd,&pkt_header))>0) {
    decode_packet(p_ptr);
    if(DEBUG) {fprintf(stdout,"\n");}
    fflush(stdout);
  }

  /* close descriptor; currently we never get here, so this is bad code :) */

  pcap_close(pd);

  return(0); /* all is well */

}

/* decode_packet ----------------------------------------------------------- */

/*
 * this code was based on discoveries done by fake & maelstrom in 1998
 *
 */

void decode_packet (char *packet) {

  int count,lc,ver;

  unsigned char byte;
  unsigned char smac[6];
  unsigned char dmac[6];
  unsigned char sip[4];
  unsigned char dip[4];

  unsigned char icmp_code;
  unsigned char icmp_type;

  struct {
    int ident;
    int seqno;
  } icmp_query,icmp_reply;

  if(DEBUG) {fprintf(stdout,"\n");}

  /* grab actual MAC address (from link layer) */

  memcpy(&dmac,&packet[0],6);
  memcpy(&smac,&packet[6],6);

  if(DEBUG) {
    for(count=0;count<6;count++) {
      fprintf(stdout,"%02x",smac[count]);
      if(count<5) 
	fprintf(stdout,":");
    }

    fprintf(stdout," => ");

    for(count=0;count<6;count++) {
      fprintf(stdout,"%02x",dmac[count]);
      if(count<5) 
	fprintf(stdout,":");
    }
  }

  /* grab some stuff from the IP header */

  memcpy(&dip,&packet[26],4);
  memcpy(&sip,&packet[30],4);

  if(DEBUG) {

    /* source address */

    fprintf(stdout," %d.%d.%d.%d -> ",
	    (unsigned char)dip[0],
	    (unsigned char)dip[1],
	    (unsigned char)dip[2],
	    (unsigned char)dip[3]);

    /* destination address */

    fprintf(stdout,"%d.%d.%d.%d ",
	    (unsigned char)sip[0],
	    (unsigned char)sip[1],
	    (unsigned char)sip[2],
	    (unsigned char)sip[3]);

  }

  /* grab useful data from ICMP payload */

  icmp_code=(unsigned char)packet[34];
  icmp_type=(unsigned char)packet[35];

  /* echo request */

  if((icmp_code == 8) && !icmp_type) {
    memcpy(&icmp_query.ident,&packet[38],2);
    memcpy(&icmp_query.seqno,&packet[40],2);
  }

  /* echo reply */

  if(!icmp_code && !icmp_type) {
    memcpy(&icmp_reply.ident,&packet[38],2);
    memcpy(&icmp_reply.seqno,&packet[40],2);
  }

  if(DEBUG) {
    fprintf(stdout,"icmp %d/%d",icmp_code,icmp_type);
  }

  /* the following loop is for debugging purposes only */

  if(DEBUG) {

    fprintf(stdout,"\n");

    lc=0;

    for(count=0;count<SNAPLEN;count++) {
      fprintf(stdout,"%02x ",(unsigned char)packet[count]);
      lc++;
      if(lc>15) {
	lc=0;
	fprintf(stdout,"\n");
      }
    }

  }
  
  /* criteria for a sessionwall-3 reply packet:
   *
   * icmp_code==0, icmp_type==0
   *
   * icmp identifier == icmp sequence number
   *
   * rversion13() || rversion14() must return 1
   *
   */

  if(!icmp_code && !icmp_type) {

    /* identifier == sequence number ? */

    if(ntohs(icmp_reply.ident) == ntohs(icmp_reply.seqno)) {

      /* version 1.3 or 1.4 ? */

      if(rversion13(smac,packet) || rversion14(smac,packet)) {

	/* sure thing, we've got a match; print out some info */

	fprintf(stdout,"\nSW-3-REPLY: v%d.%d detected at MAC ",
		(unsigned char)packet[43],(unsigned char)packet[42]);

	for(count=0;count<6;count++) {
	  fprintf(stdout,"%02x",smac[count]);
	  if(count<5) 
	    fprintf(stdout,":");
	}

      }

    }

  }

  /* criteria for a sessionwall-3 discover packet:
   *
   * icmp_code==8, icmp_type==0
   *
   * icmp identifier = icmp sequence number
   *
   * (possibly also 4 bytes after sequence number must be zero)
   *
   * dversion13() || dversion14() must return > 0
   *
   */

  if((icmp_code==8) && !icmp_type) {

    /* identifier == sequence number ? */

    if(ntohs(icmp_query.ident) == ntohs(icmp_query.seqno)) {

      /* version 1.3 or 1.4 ? */

      if((ver=dversion13(smac,packet)) || (ver=dversion14(smac,packet))) {

	/* sure thing, we've got a match; print out some info (this could
	 * be fake!!) 
	 */

	if(ver<14) {
	  fprintf(stdout,"\nSW-3-DISCOVER: v1.3 or LOWER detected at MAC ",ver);
	} else {
	  fprintf(stdout,"\nSW-3-DISCOVER: v1.4 or HIGHER detected at MAC ",ver);
	}

	for(count=0;count<6;count++) {
	  fprintf(stdout,"%02x",smac[count]);
	  if(count<5) 
	    fprintf(stdout,":");
	}

      }

    }

  }


}

/* dversion13 -------------------------------------------------------------- */
/* returns 1 if discover packet from version 1.3 (or less?) is detected
 * UNTESTED!!
 */

int dversion13(char *mac, char *pkt) {
  if(((unsigned char)pkt[45] == ((unsigned char)mac[0] ^ 0x44)) &&
     ((unsigned char)pkt[49] == ((unsigned char)mac[1] ^ 0x41)) &&
     ((unsigned char)pkt[61] == ((unsigned char)mac[2] ^ 0x73)) &&
     ((unsigned char)pkt[55] == ((unsigned char)mac[3] ^ 0x43)) &&
     ((unsigned char)pkt[69] == ((unsigned char)mac[4] ^ 0x6c)) &&
     ((unsigned char)pkt[57] == ((unsigned char)mac[5] ^ 0x55))) {
    return(13);
  }
  return(0);
}

/* dversion14 -------------------------------------------------------------- */
/* returns 1 if discover packet from version 1.4 is detected */

int dversion14(char *mac, char *pkt) {
  if(((unsigned char)pkt[46] == ((unsigned char)mac[0] ^ 0x44)) &&
     ((unsigned char)pkt[51] == ((unsigned char)mac[1] ^ 0x41)) &&
     ((unsigned char)pkt[62] == ((unsigned char)mac[2] ^ 0x73)) &&
     ((unsigned char)pkt[54] == ((unsigned char)mac[3] ^ 0x43)) &&
     ((unsigned char)pkt[67] == ((unsigned char)mac[4] ^ 0x6c)) &&
     ((unsigned char)pkt[60] == ((unsigned char)mac[5] ^ 0x55))) {
    return(14);
  }
  return(0);
}

/* rversion13 -------------------------------------------------------------- */
/* returns 1 if reply packet from version 1.3 (or less?) is detected 
 * UNTESTED!!
 */

int rversion13(char *mac, char *pkt) {
  if(((unsigned char)pkt[45] == ((unsigned char)mac[0] ^ 0x4b)) &&
     ((unsigned char)pkt[49] == ((unsigned char)mac[1] ^ 0x6f)) &&
     ((unsigned char)pkt[61] == ((unsigned char)mac[2] ^ 0x53)) &&
     ((unsigned char)pkt[55] == ((unsigned char)mac[3] ^ 0x61)) &&
     ((unsigned char)pkt[69] == ((unsigned char)mac[4] ^ 0x4d)) &&
     ((unsigned char)pkt[57] == ((unsigned char)mac[5] ^ 0x72))) {
    return(1);
  }
  return(0);
}

/* rversion14 -------------------------------------------------------------- */
/* returns 1 if reply packet from version 1.4 is detected */

int rversion14(char *mac, char *pkt) {
  if(((unsigned char)pkt[46] == ((unsigned char)mac[0] ^ 0x4b)) &&
     ((unsigned char)pkt[51] == ((unsigned char)mac[1] ^ 0x6f)) &&
     ((unsigned char)pkt[62] == ((unsigned char)mac[2] ^ 0x53)) &&
     ((unsigned char)pkt[54] == ((unsigned char)mac[3] ^ 0x61)) &&
     ((unsigned char)pkt[67] == ((unsigned char)mac[4] ^ 0x4d)) &&
     ((unsigned char)pkt[60] == ((unsigned char)mac[5] ^ 0x72))) {
    return(1);
  }
  return(0);
}





