/* sw3spoof.c
 * spoofs a sw-3 discovery packet
 * v1.00 17/05/2000 (c) codex@bogus.net
 *
 * unable to hack this without maelstrom and fake, their discoveries were
 * _vital_ to this project.
 *
 * this code is dependent on libnet; get it at www.packetfactory.net
 *
 * some of this code was gratuituously ripped from libnet-sample.c
 * written by mike@infonexus.com and route@infonexus.com; i am sure they
 * won't mind.
 */

#define ERRBUFSIZE 256
#define SNAPLEN 74
#define SW3_PACKET_LEN 52           /* 52+ETH_H+ICMP_H == 74 */

#include <libnet.h>

unsigned char b;
unsigned char PcA;
unsigned char PcB;
unsigned char NcA;
unsigned char NcB;

unsigned char buf[SNAPLEN];
unsigned char payload[SW3_PACKET_LEN];

u_char src_mac[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
u_char dst_mac[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff}; 

/* rnd() -- generate a random number between 0x00 and 0xff ----------------- */

int rnd(void) {
  return(rand()%0xff);
}

/* usage() ----------------------------------------------------------------- */

void usage(void) {
  fprintf(stdout,"sw3spoof v1.00 17/05/2000 (c) codex@bogus.net\n");
  fprintf(stdout,"usage: sw3spoof [-c <count>] [-h] [-r] [-s <source>] -i <interface> -v <13|14> -t <target>\n");
  fprintf(stdout,"\t -h : this\n");
  fprintf(stdout,"\t -r : send reply instead of discovery packet\n");
  fprintf(stdout,"\t -i : network device\n");
  fprintf(stdout,"\t -v : version number 13 or 14\n");
  fprintf(stdout,"\t -t : target (usually a broadcast address)\n");
  fprintf(stdout,"\t -s : source address (defaults to random)\n");
  fprintf(stdout,"\t -c : number of packets (default 1)\n");
}

/* spoof_sw3discovery() -- construct fake packet and send it --------------- */

int spoof_sw3discovery(struct libnet_link_int *link, unsigned long src_ip, 
		       unsigned long dst_ip, unsigned char *device,
		       int version) {

  int sent;
 
  unsigned short ident,seqno;

  ident=rand()%65535;
  seqno=ident;

  bzero(payload,sizeof(payload));

  /* build ethernet header; fill the source MAC with some random crap */
  
  src_mac[0]=rnd();
  src_mac[1]=rnd();
  src_mac[2]=rnd();
  src_mac[3]=rnd();
  src_mac[4]=rnd();
  src_mac[5]=rnd(); 

  libnet_build_ethernet(dst_mac,src_mac,ETHERTYPE_IP,NULL,SNAPLEN,buf);
  
  /* build something which resembles a valid ip header */

  if(!src_ip) src_ip=rnd();

  libnet_build_ip(40,0,rnd(),0,0x20,IPPROTO_ICMP,src_ip,dst_ip,
		  NULL,0,buf+LIBNET_ETH_H);
  
  /* fill some "valid" sw-3 data into the buffer -- we currently only have
   * valid info on version v1.3 or less and v1.4. ok, a lame way of doing it
   * but it works.
   */

  if(version!=13 && version!=14) {
    fprintf(stderr,"error: unknown version %d\n",version);
    exit(1);
  }

  if(version==14) {
    (unsigned char)payload[4] = ((unsigned char)src_mac[0] ^ 0x44);
    (unsigned char)payload[9] = ((unsigned char)src_mac[1] ^ 0x41);
    (unsigned char)payload[20] = ((unsigned char)src_mac[2] ^ 0x73);
    (unsigned char)payload[12] = ((unsigned char)src_mac[3] ^ 0x43);
    (unsigned char)payload[25] = ((unsigned char)src_mac[4] ^ 0x6c);
    (unsigned char)payload[18] = ((unsigned char)src_mac[5] ^ 0x55);
  }

  if(version==13) {
    (unsigned char)payload[3] = ((unsigned char)src_mac[0] ^ 0x44);
    (unsigned char)payload[7] = ((unsigned char)src_mac[1] ^ 0x41);
    (unsigned char)payload[19] = ((unsigned char)src_mac[2] ^ 0x73);
    (unsigned char)payload[13] = ((unsigned char)src_mac[3] ^ 0x43);
    (unsigned char)payload[27] = ((unsigned char)src_mac[4] ^ 0x6c);
    (unsigned char)payload[15] = ((unsigned char)src_mac[5] ^ 0x55);
  }

  libnet_build_icmp_echo(ICMP_ECHO,0,ident,seqno,payload,
     SW3_PACKET_LEN,buf+LIBNET_ETH_H+LIBNET_IP_H);

  /* calculate checksum */

  if((libnet_do_checksum(buf+LIBNET_ETH_H,IPPROTO_ICMP,
			 SW3_PACKET_LEN+LIBNET_ICMP_ECHO_H))<0) {
    fprintf(stderr,"error: libnet_do_checksum: failed on icmp\n");
 
  }

  if((libnet_do_checksum(buf+LIBNET_ETH_H,IPPROTO_IP,
			 SW3_PACKET_LEN+LIBNET_IP_H+LIBNET_ICMP_ECHO_H))<0) {
    fprintf(stderr,"error: libnet_do_checksum: failed on ip\n");
  }

  /* send data -- a sw-3 discovery packet is 74 bytes */

  sent=libnet_write_link_layer(link,device,buf,SNAPLEN);

  return(sent);

}

/* sw3reply() --- construct a fake reply packet and send it ---------------- */

int spoof_sw3reply(struct libnet_link_int *link, unsigned long src_ip, 
		   unsigned long dst_ip,unsigned char *device, int version) {
  int c,sent;
 
  unsigned short ident,seqno;

  ident=rand()%65535;
  seqno=ident;

  bzero(payload,sizeof(payload));

  /* build ethernet header; fill the source MAC with some random crap */
  
  memcpy(dst_mac,src_mac,6);

  src_mac[0]=rnd();
  src_mac[1]=rnd();
  src_mac[2]=rnd();
  src_mac[3]=rnd();
  src_mac[4]=rnd();
  src_mac[5]=rnd();

  libnet_build_ethernet(dst_mac,src_mac,ETHERTYPE_IP,NULL,SNAPLEN,buf);
  
  /* build something which resembles a valid ip header */

  if(!src_ip) src_ip=rnd();

  libnet_build_ip(40,0,rnd(),0,0x20,IPPROTO_ICMP,src_ip,dst_ip,
                  NULL,0,buf+LIBNET_ETH_H);
  
  /* fill some "valid" sw-3 data into the buffer -- we currently only have
   * valid info on version v1.3 or less and v1.4. ok, a lame way of doing it
   * but it works.
   */

  if(version!=13 && version!=14) {
    fprintf(stderr,"error: unknown version (use 13 or 14) %d\n",version);
    exit(1);
  }

  if(version==14) {
    (unsigned char)payload[4] = ((unsigned char)src_mac[0] ^ 0x4b);
    (unsigned char)payload[9] = ((unsigned char)src_mac[1] ^ 0x6f);
    (unsigned char)payload[20] = ((unsigned char)src_mac[2] ^ 0x53);
    (unsigned char)payload[12] = ((unsigned char)src_mac[3] ^ 0x61);
    (unsigned char)payload[25] = ((unsigned char)src_mac[4] ^ 0x4d);
    (unsigned char)payload[18] = ((unsigned char)src_mac[5] ^ 0x72);
    (unsigned char)payload[0] = 4;
    (unsigned char)payload[1] = 1;
  }

  if(version==13) {
    (unsigned char)payload[3] = ((unsigned char)src_mac[0] ^ 0x4b);
    (unsigned char)payload[7] = ((unsigned char)src_mac[1] ^ 0x6f);
    (unsigned char)payload[19] = ((unsigned char)src_mac[2] ^ 0x53);
    (unsigned char)payload[13] = ((unsigned char)src_mac[3] ^ 0x61);
    (unsigned char)payload[27] = ((unsigned char)src_mac[4] ^ 0x4d);
    (unsigned char)payload[15] = ((unsigned char)src_mac[5] ^ 0x72);
    (unsigned char)payload[0] = 3;
    (unsigned char)payload[1] = 1;
  }

  libnet_build_icmp_echo(ICMP_ECHOREPLY,0,ident,seqno,payload,
     SW3_PACKET_LEN,buf+LIBNET_ETH_H+LIBNET_IP_H);

  /* whack some extra data into the packet */

  PcA++;
  if(PcA==0x00) PcB++;

  NcA=NcA-2;

  if(NcA==0x00) {
    c=rand()%100;
    fprintf(stderr,"c==%d\n",c);
    if(c<34) NcB=b;
    if(c>33 && c<67) NcB=(b-1);
    if(c>66) NcB=(b-2);
  }

  (unsigned char)buf[18]=PcA;  /* fast positive 8bit counter */
  (unsigned char)buf[19]=PcB;  /* slow positive 8bit counter */
  (unsigned char)buf[36]=NcA;  /* fast negative 8bit counter */
  (unsigned char)buf[37]=NcB;  /* slow negative 8bit counter */

  (unsigned char)buf[38]=(PcA+1);
  (unsigned char)buf[39]=0;
  (unsigned char)buf[40]=(PcA+1);
  (unsigned char)buf[41]=0;

  /* calculate a checksum, thanks to fake for pointing out the order */

  if((libnet_do_checksum(buf+LIBNET_ETH_H,IPPROTO_ICMP,
                         SW3_PACKET_LEN+LIBNET_ICMP_ECHO_H))<0) {
    fprintf(stderr,"error: libnet_do_checksum: failed on icmp\n");
 
  }

  if((libnet_do_checksum(buf+LIBNET_ETH_H,IPPROTO_IP,
                         SW3_PACKET_LEN+LIBNET_IP_H+LIBNET_ICMP_ECHO_H))<0) {
    fprintf(stderr,"error: libnet_do_checksum: failed on ip\n");
  }

  /* send data -- a sw-3 discovery packet is 74 bytes */

  sent=libnet_write_link_layer(link,device,buf,SNAPLEN);

  return(sent);

}

/* the big enchilada ------------------------------------------------------- */

int main (int argc, char *argv[]) {
  int c,sent,version,count;
  int discover=1;
  char errbuf[ERRBUFSIZE];
  char *device = NULL;
  unsigned long dst_ip=0;
  unsigned long src_ip=0;
  struct libnet_link_int *link;

  srand(getpid()); 

  /* what does the user really want?? get some clue from the command line */

  while((c = getopt(argc,argv,"c:v:i:s:t:hr")) != EOF) {
    switch(c) {
    case 'c':
      count=atoi(optarg);
      break;
    case 'i':
      device=optarg;
      break;
    case 'v':
      version=atoi(optarg);
      break;
    case 't':
      if ((dst_ip = libnet_name_resolve(optarg, 1)) == -1) {
	fprintf(stderr,"error: can't resolve host '%s'\n",optarg);
	exit(1);
      }
      break;
    case 's':
      if ((src_ip = libnet_name_resolve(optarg, 1)) == -1) {
	fprintf(stderr,"error: can't resolve host '%s'\n",optarg);
	exit(1);
      }
      break;
    case 'r':
      discover=0;
      break;
    case 'h':
      usage();
      exit(0);
    default:
      usage();
      exit(1);
    }
  }

  if(!count) count=1;

  /* bail out if no destination ip or device is given */

  if(!device || !dst_ip) {
    usage();
    exit(1);
  }

  if(!(link = libnet_open_link_interface(device,errbuf))) {
    fprintf(stderr,"error: libnet_open_link_interface: %s\n",errbuf);
    exit(1);
  }
  
  /* this is where we want to do the huge doorknob action to the pods */

  for(c=0;c<count;c++) {
    if(discover) {
      sent=spoof_sw3discovery(link,src_ip,dst_ip,device,version);
      if(sent<0) break;
    } else {
      sent=spoof_sw3reply(link,src_ip,dst_ip,device,version);
      if(sent<0) break;
    }
  }

  /* complain if we couldn't send any data */

  if(sent<0) {
    fprintf(stderr,"error: libnet_write_link_layer returned %d\n", sent);
    exit(1);
  }

  return(0);
}















